from datetime import (
    datetime,
    timedelta,
)
from django.contrib.humanize.templatetags.humanize import naturaltime
from django.utils.timezone import get_current_timezone
from django.forms.utils import pretty_name
from django.contrib.admin.utils import FieldIsAForeignKeyColumnName
from django.core.exceptions import (
    FieldDoesNotExist,
)


class PassThroughWriter:
    """ Can be used with StreamingHttpResponse to avoid buffering. """
    def write(self, value):
        return value


def init_kwargs(model, arg_dict):
    model_fields = [f.name for f in model._meta.get_fields()]
    return {k: v for k, v in arg_dict.items() if k in model_fields}


def label_for_field(name, model, view=None, return_attr=False):
    """
    Return a sensible label for a field name. The name can be a callable,
    property (but not created with @property decorator), or the name of an
    object's attribute, as well as a model field. If return_attr is True, also
    return the resolved attribute (which could be a callable). This will be
    None if (and only if) the name refers to a field.
    """
    attr = None
    if '__' in name:
        names = name.split('__')
        model_names = names[:-1]
        for name in model_names:
            model = model._meta.get_field(name).related_model

        field_name = names[-1]
        related_field = model._meta.get_field(field_name)

        label = related_field.verbose_name
        attr = str
    else:
        try:
            field = model._meta.get_field(name)
            try:
                label = field.verbose_name
            except AttributeError:
                # field is likely a ForeignObjectRel
                label = field.related_model._meta.verbose_name
        except FieldDoesNotExist:
            if name == "__str__":
                label = str(model._meta.verbose_name)
                attr = str
            else:
                if callable(name):
                    attr = name
                elif hasattr(view, name):
                    attr = getattr(view, name)
                elif hasattr(model, name):
                    attr = getattr(model, name)
                else:
                    label = name

                if hasattr(attr, "short_description"):
                    label = attr.short_description
                elif (isinstance(attr, property) and
                      hasattr(attr, "fget") and
                      hasattr(attr.fget, "short_description")):
                    label = attr.fget.short_description
                elif callable(attr):
                    if attr.__name__ == "<lambda>":
                        label = "--"
                    else:
                        label = pretty_name(attr.__name__)
                else:
                    label = pretty_name(name)
        except FieldIsAForeignKeyColumnName:
            label = pretty_name(name)
            attr = name

    if return_attr:
        return (label, attr)
    else:
        return label


def format_datetime(d):
    return f'{d.strftime("%Y-%m-%d %H:%M")} ({naturaltime(d)})'



from django import template

register = template.Library()


@register.inclusion_tag('datatables/css.html')
def css_bundle():
    return {}


@register.inclusion_tag('datatables/js.html')
def js_bundle():
    return {}



import csv
from datetime import datetime
from django.http import (
    JsonResponse,
    StreamingHttpResponse,
)
from django.views.generic import TemplateView
from .datatable import Datatable
from .utils import PassThroughWriter


class DatatableView(TemplateView):
    template_name = 'datatables/datatable.html'
    model = None
    columns = None
    list_filters = None
    filter_lookup = None
    ordering = None
    search_fields = None
    table_search_fields = None
    table_config = None
    lookup_opts = None
    filters = None
    exportable = False
    update_interval = 60000  # Milliseconds

    def dispatch(self, request, *args, **kwargs):
        self.init_datatable(request, **kwargs)
        return super().dispatch(request, *args, **kwargs)

    def init_datatable(self, request, **kwargs):
        self.datatable = Datatable(
            request,
            self.model,
            self,
            self.get_columns(),
            self.get_update_interval(),
            self.get_list_filters(),
            self.get_filter_lookup(),
            self.get_ordering(),
            self.get_search_fields(),
            self.get_table_search_fields(),
            self.get_table_config(**kwargs),
        )
        if self.lookup_opts:
            self.datatable.lookup_opts = self.lookup_opts
        return self.datatable

    def get_columns(self):
        return self.columns or []

    def get_update_interval(self):
        return self.update_interval

    def get_export_columns(self):
        return self.request.GET.getlist('visible[]')

    def get_list_filters(self):
        return self.list_filters or []

    def get_filter_lookup(self):
        return self.filter_lookup or {}

    def get_ordering(self):
        return self.ordering or []

    def get_filters(self):
        return self.filters or []

    def get_search_fields(self):
        return self.search_fields or []

    def get_table_search_fields(self):
        return self.table_search_fields or []

    def get_table_config(self, **kwargs):
        return self.table_config or {}

    def get_queryset(self):
        qs = self.model.objects.all()
        return qs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['main_search'] = self.request.GET.get('main_search', '')
        context['datatable'] = self.datatable
        return context

    def render_to_json_response(self):
        data = self.get_data()
        return JsonResponse(self.datatable.serialize(data))

    def download_as_csv(self):
        def csv_row_generator(query_set):
            writer = csv.writer(PassThroughWriter())
            headers = [c[1] if isinstance(c, tuple) else c for c in self.get_export_columns()]

            yield writer.writerow(headers)

            for obj in query_set:
                row = [self.datatable.get_column_value(obj, c, raw=True) for c in self.get_export_columns()]
                yield writer.writerow(row)

        query_set = self.get_data()
        response = StreamingHttpResponse(csv_row_generator(query_set), content_type='text/csv')
        filename = self.model._meta.verbose_name_plural + "_{:%Y-%m-%dT%H-%M}".format(datetime.now())
        filename = filename.replace(' ', '_')
        response['Content-Disposition'] = f'attachment; filename="{filename}.csv"'
        return response

    def get_data(self):
        return self.datatable.get_queryset()

    def get(self, request, *args, **kwargs):
        if request.GET.get('format') == 'json':
            return self.render_to_json_response()
        elif request.GET.get('format') == 'csv':
            return self.download_as_csv()
        else:
            context = self.get_context_data()
            return self.render_to_response(context)



from urllib.parse import urlencode

from django.contrib.admin.utils import get_model_from_relation
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import gettext_lazy as _


class ListFilter:
    title = None  # Human-readable title to appear in the right sidebar.
    template = 'admin/filter.html'

    def __init__(self, model):
        # This dictionary will eventually contain the request's query string
        # parameters actually used by this filter.
        if self.title is None:
            raise ImproperlyConfigured(
                "The list filter '%s' does not specify a 'title'."
                % self.__class__.__name__
            )

    def has_output(self):
        """
        Return True if some choices would be output for this filter.
        """
        raise NotImplementedError('subclasses of ListFilter must provide a has_output() method')

    def choices(self, changelist):
        """
        Return choices ready to be output in the template.
        `changelist` is the ChangeList to be displayed.
        """
        raise NotImplementedError('subclasses of ListFilter must provide a choices() method')

    def expected_parameters(self):
        """
        Return the list of parameter names that are expected from the
        request's query string and that will be used by this filter.
        """
        raise NotImplementedError('subclasses of ListFilter must provide an expected_parameters() method')


class FieldListFilter(ListFilter):
    _field_list_filters = []
    _take_priority_index = 0

    def __init__(self, field, model, field_path):
        self.field = field
        self.field_path = field_path
        self.title = getattr(field, 'verbose_name', field_path)
        # If the field has a verbose_name field but no value
        # capitalize the field_path
        if self.title is None:
            self.title = field_path.capitalize()
        super().__init__(model)

    def has_output(self):
        return True

    @classmethod
    def register(cls, test, list_filter_class, take_priority=False):
        if take_priority:
            # This is to allow overriding the default filters for certain types
            # of fields with some custom filters. The first found in the list
            # is used in priority.
            cls._field_list_filters.insert(
                cls._take_priority_index, (test, list_filter_class))
            cls._take_priority_index += 1
        else:
            cls._field_list_filters.append((test, list_filter_class))

    @classmethod
    def create(cls, field, model, field_path, **settings):
        for test, list_filter_class in cls._field_list_filters:
            if test(field):
                return list_filter_class(
                    field, model, field_path=field_path, **settings)


class RelatedFieldListFilter(FieldListFilter):
    def __init__(self, field, model, field_path, **settings):
        other_model = get_model_from_relation(field)
        self.lookup_kwarg = '%s__%s__in' % (field_path, field.target_field.name)
        self.lookup_kwarg_isnull = '%s__in' % field_path
        super().__init__(field, model, field_path)

        if settings.get('choices'):
            self.lookup_choices = settings.get('choices')
        else:
            self.lookup_choices = self.field_choices(field)

        if settings.get('title'):
            self.title = settings.get('title')
        elif hasattr(field, 'verbose_name'):
            self.title = field.verbose_name.capitalize()
        else:
            self.title = other_model._meta.verbose_name.capitalize()

    def has_output(self):
        return len(self.lookup_choices) > 1

    def expected_parameters(self):
        return [self.lookup_kwarg, self.lookup_kwarg_isnull]

    def field_choices(self, field):
        return field.get_choices(include_blank=False)

    def choices(self):
        for pk_val, val in self.lookup_choices:
            yield {
                'query_string': urlencode({self.lookup_kwarg: pk_val}),
                'display': val,
            }


FieldListFilter.register(lambda f: f.remote_field, RelatedFieldListFilter)


class BooleanFieldListFilter(FieldListFilter):
    def __init__(self, field, model, field_path, **settings):
        self.lookup_kwarg = '%s__in' % field_path
        super().__init__(field, model, field_path)

        if settings.get('choices'):
            self.lookup_choices = settings.get('choices')
        else:
            self.lookup_choices = (
                ('1', _('Yes')),
                ('0', _('No')))

        if settings.get('title'):
            self.title = settings.get('title')
        elif hasattr(field, 'verbose_name'):
            self.title = field.verbose_name.capitalize()

    def expected_parameters(self):
        return [self.lookup_kwarg]

    def choices(self):
        for lookup, title in self.lookup_choices:
            yield {
                'query_string': urlencode({self.lookup_kwarg: lookup}),
                'display': title,
            }
        if self.field.null:
            yield {
                'query_string': urlencode({self.lookup_kwarg2: 'True'}),
                'display': _('Unknown'),
            }


FieldListFilter.register(lambda f: isinstance(f, models.BooleanField), BooleanFieldListFilter)


class ChoicesFieldListFilter(FieldListFilter):
    def __init__(self, field, model, field_path, **settings):
        self.lookup_kwarg = '%s__in' % field_path
        super().__init__(field, model, field_path)
        self.checked = []

        if settings.get('choices'):
            self.lookup_choices = settings.get('choices')
        else:
            self.lookup_choices = self.field.flatchoices

        if settings.get('title'):
            self.title = settings.get('title')
        elif hasattr(field, 'verbose_name'):
            self.title = field.verbose_name.capitalize()

        if settings.get('checked'):
            self.checked = settings.get('checked')

        self.has_grouped_choices = settings.get('grouped_choices', False)

    def expected_parameters(self):
        return [self.lookup_kwarg, self.lookup_kwarg_isnull]

    def choices(self, choices=None):
        if not choices:
            choices = self.lookup_choices

        for lookup, title in choices:
            if lookup is None:
                continue

            yield {
                'query_string': urlencode({self.lookup_kwarg: lookup}),
                'display': title,
                'checked': lookup in self.checked,
                'val': lookup,
            }

    def grouped_choices(self):
        for group, choices in self.lookup_choices:
            yield {
                'display': group,
                'checked': group in self.checked,
            }, self.choices(choices)


FieldListFilter.register(lambda f: bool(f.choices), ChoicesFieldListFilter)
FieldListFilter.register(lambda f: isinstance(f, models.CharField), ChoicesFieldListFilter)

try:
    from django.contrib.postgres.fields import ArrayField
except ModuleNotFoundError:
    pass
else:
    class ArrayFieldListFilter(FieldListFilter):
        def __init__(self, field, model, field_path, **settings):
            self.lookup_kwarg = '%s__overlap' % field_path
            self.lookup_kwarg_isnull = '%s__len' % field_path
            super().__init__(field, model, field_path)

            if settings.get('choices'):
                self.lookup_choices = settings.get('choices')
            else:
                self.lookup_choices = [(f'{choice}', title) for choice, title
                                       in self.field.base_field.flatchoices]

            self.empty_value_display = 'None'

            if settings.get('title'):
                self.title = settings.get('title')
            elif hasattr(field, 'verbose_name'):
                self.title = field.verbose_name.capitalize()

        def expected_parameters(self):
            return [self.lookup_kwarg, self.lookup_kwarg_isnull]

        def choices(self):
            for lookup, title in self.lookup_choices:
                if lookup is None:
                    continue
                yield {
                    'query_string': urlencode({self.lookup_kwarg: lookup}),
                    'display': title,
                }

            yield {
                'query_string': urlencode({self.lookup_kwarg_isnull: '0'}),
                'display': self.empty_value_display,
            }


    FieldListFilter.register(lambda f: isinstance(f, ArrayField), ArrayFieldListFilter)


class DateTimeFieldListFilter(FieldListFilter):
    def __init__(self, field, model, field_path, **settings):
        super().__init__(field, model, field_path)
        self.lookup_kwarg = '%s__' % field_path
        self._endpoints = settings.get('endpoints')

        if settings.get('title'):
            self.title = settings.get('title')
        elif hasattr(field, 'verbose_name'):
            self.title = field.verbose_name.capitalize()
        self.is_date = True

    @staticmethod
    def valid_predicates():
        return ['gte', 'lte', 'gt', 'lt']

    @classmethod
    def contains_valid_predicate(cls, f):
        return any(i for i in cls.valid_predicates() if i in f)

    def endpoints(self):
        for endpoint in self._endpoints:
            yield {
                'lookup': self.lookup_kwarg + endpoint['predicate'],
                'display': endpoint['label']
            }

    def choices(self, changelist):
        yield {}

    def expected_parameters(self):
        return []


FieldListFilter.register(lambda f: isinstance(f, models.DateField), DateTimeFieldListFilter)
FieldListFilter.register(lambda f: isinstance(f, models.DateTimeField), DateTimeFieldListFilter)



#!/usr/bin/env python
import os

from setuptools import find_packages, setup

setup(
    name='django-datatablesview',
    version='0.0.8',
    description='A Django app that integrates with DataTables.net javascript api.',
    author='dahlkar',
    url='http://github.com/dahlkar/django-datatables',
    packages=find_packages(exclude=['example', 'example.*']),
    install_requires=[
        "django>=2.1"
    ],
    include_package_data=True,
    zip_safe=True,
    license='MIT License',
    classifiers=[
        "Programming Language :: Python",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
    ]
)



import json

from datetime import datetime

from django.template.loader import render_to_string
from django.contrib.admin.utils import get_fields_from_path
from django.db import models
from django.core.paginator import Paginator
from django.utils.timezone import get_current_timezone

from .datatable_filters import FieldListFilter, DateTimeFieldListFilter
from .utils import (
    label_for_field,
    format_datetime,
)


class Datatable:
    template = 'datatables/_datatable.html'
    table_template = 'datatables/table.html'
    filter_template = 'datatables/filters.html'
    export_template = 'datatables/export.html'
    toggle_template = 'datatables/toggle-columns.html'
    config = {}

    def __init__(
            self,
            request,
            model,
            view,
            columns,
            update_interval,
            filters=[],
            filter_lookup=[],
            ordering=[],
            search_fields=[],
            table_search_fields=[],
            config={},
            show_columns=True,
    ):
        self.request = request
        self.model = model
        self.view = view
        self.columns = columns
        self._filters = filters
        self.filter_lookup = filter_lookup
        self.lookup_opts = []
        if model:
            self.lookup_opts = [f.name for f in model._meta.get_fields()]

        self.filter_specs = self.get_filters()
        self.ordering = ordering
        self.search_fields = search_fields
        self.table_search_fields = table_search_fields
        self.config = config
        self.show_columns = show_columns
        self.update_interval = update_interval
        self.exportable = view.exportable

    def __str__(self):
        return render_to_string(self.template, self.get_context())

    def get_context(self):
        return {
            'id': self.get_id(),
            'datatable': self,
            'columns': self.get_columns(),
            'config': self.get_config(),
        }

    def table(self):
        return render_to_string(self.table_template, self.get_context())

    def filters(self):
        return render_to_string(self.filter_template, self.get_context())

    def export_button(self):
        return render_to_string(self.export_template, self.get_context())

    def toggle_columns(self):
        return render_to_string(self.toggle_template, self.get_context())

    def get_id(self):
        return id(self)

    def get_columns(self):
        return self.get_table_columns()

    def get_config(self):
        return json.dumps(self.config)

    def get_filters(self):
        self.filter_specs = {}
        for list_filter, *settings in self._filters:
            try:
                settings = dict(settings)
            except ValueError:
                settings = {}

            field_path = None
            field, field_list_filter_class = list_filter, FieldListFilter.create
            if not isinstance(field, models.Field):
                # For annotated data you have to specify the type
                if 'field_type' in settings:
                    # No path for annotated fields
                    field_path = list_filter
                    field = settings['field_type']()
                else:
                    field_path = field
                    field = get_fields_from_path(self.model, field_path)[-1]

            spec = field_list_filter_class(
                field, self.model, field_path=field_path,
                **settings,
            )
            if spec and spec.has_output():
                self.filter_specs[list_filter] = spec

        return self.filter_specs.values()

    def serialize(self, qs):
        page, page_size = self.parse_page_info()
        pager = Paginator(qs, page_size)
        return {
            "recordsTotal": self.total,
            "recordsFiltered": pager.count,
            "data": self.serialize_data(pager.page(page)),
        }

    def serialize_data(self, data):
        result = []
        for item in data:
            obj = {}
            columns = self.columns
            for column in columns:
                if isinstance(column, tuple):
                    column = column[0]

                obj[column] = self.get_column_value(item, column)

            result.append(obj)

        return result

    def get_column_value(self, item, column, raw=False):
        if isinstance(column, tuple):
            column = column[0]
        if hasattr(self.view, column):
            view_func = getattr(self.view, column)
            attr = view_func(item)
        elif '__' in column:
            fields = column.split("__")
            attr = item
            for field in fields:
                attr = self._get_attr(attr, field)
        else:
            attr = self._get_attr(item, column)

        if isinstance(attr, datetime):
            attr = attr.astimezone(get_current_timezone())
            if raw:
                attr = f'{attr.strftime("%Y-%m-%d %H:%M")}'
            else:
                attr = format_datetime(attr)

        if attr is None:
            attr = '-'

        return str(attr)

    def _get_attr(self, item, column):
        if hasattr(item, f'get_{column}_display'):
            attr = getattr(item, f'get_{column}_display')()
        elif hasattr(item, column):
            attr = getattr(item, column)
        else:
            attr = ''

        return attr

    def get_orderable_columns(self):
        available = self.lookup_opts
        if self.model:
            available += list(self.view.get_queryset().query.annotation_select)
        return available

    def get_queryset(self):
        qs = self.view.get_queryset()
        self.total = qs.count()
        filters = self.parse_filters()

        filters &= self.parse_search()
        qs = qs.filter(filters)
        ordering = self.parse_ordering()
        qs = qs.order_by(*ordering)
        return qs.distinct()

    def _create_search_query(self, fields, search):
        q = models.Q()
        for search_field in fields:
            q |= models.Q(**{f'{search_field}__icontains': search})

        return q

    def parse_search(self):
        search = self.request.GET.get('main_search')
        table_search = self.request.GET.get('search[value]')
        query = models.Q()
        if search:
            query &= self._create_search_query(self.search_fields, search)
        if table_search:
            query &= self._create_search_query(self.search_fields, table_search)
        return query

    def parse_filters(self):
        filters = [i.split("=") for i in self.request.GET.getlist('filters[]') if len(i.split("=")) == 2]
        args = {}
        for f, a in filters:
            if '__in' in f or '__overlap' in f:
                tmp = args.get(f, [])
                if a in self.filter_lookup:
                    a = self.filter_lookup.get(a)
                    tmp.extend(a)
                else:
                    tmp.append(a)
                args[f] = tmp
            elif DateTimeFieldListFilter.contains_valid_predicate(f):
                date, time = a.split(' ')
                year, month, day = date.split('-')
                hour, _ = time.split(':')
                args[f] = datetime(int(year), int(month), int(day), int(hour))
            else:
                args[f] = a

        query = models.Q()
        for f in args:
            query &= models.Q(**{f: args[f]})

        for f in self.view.get_filters():
            query &= models.Q(**f)

        return query

    def parse_page_info(self):
        page_size = int(self.request.GET.get("length", 25))
        page = int(self.request.GET.get("start", 0)) / page_size
        return (page + 1, page_size)

    def parse_ordering(self):
        sort_cols = self.ordering
        for i in range(0, len(self.columns)):
            sort_col = self.request.GET.get(f'order[{i}][column]')
            if sort_col is None:
                return sort_cols
            sort_col = int(sort_col)
            col = self.columns[sort_col]
            if isinstance(col, tuple):
                col = col[0]
            sort_dir = '-' if self.request.GET.get(f'order[{i}][dir]') == 'desc' else ''
            sort_col_name = sort_dir + col
            sort_cols.append(sort_col_name)

        return sort_cols

    def get_table_columns(self):
        """
        Generate the list column headers.
        """
        columns = []
        orderable = self.get_orderable_columns()
        for field_name in self.columns:
            visible = True
            togglable = False
            if isinstance(field_name, tuple):
                try:
                    settings = field_name[2]
                    visible = settings.get('visible', True)
                    togglable = settings.get('togglable', False)
                except IndexError:
                    pass
                label = field_name[1]
                field_name = field_name[0]
            else:
                label = label_for_field(field_name, self.model, self.view)

            columns.append((label, field_name, field_name in orderable, visible, togglable))

        return columns
